/**
 * Constants
 */
var gc = gc || {};

// packet types
let COMMAND_PACKET = 1;
let REPLY_PACKET = 2;
let ERROR_PACKET = 3;
let PAYLOAD_PACKET = 4;
let INTERRUPT_PACKET = 5;

// packet flags
let FLAG_MOREDATA = 1;

let PACKET_IDENTIFIER = "T".charCodeAt(0);

let PACKET_ID = 0;
let PACKET_PEC = 1;
let PACKET_PAYLOAD_LEN = 2;
let PACKET_TYPE = 3;
let PACKET_FLAGS = 4;
let PACKET_SEQ_NUM = 5;
let PACKET_STATUS = 6;
let PACKET_COMMAND = 7;
let PACKET_PAYLOAD = 8;

let MAX_PACKET_SIZE = 62; // size, in bytes, of a USB packet
let PACKET_HEADER_SIZE = PACKET_PAYLOAD;
let MAX_PAYLOAD = MAX_PACKET_SIZE - PACKET_HEADER_SIZE;

let Command = {
  Cmd_I2C_Control: 1, // 0x01
  Cmd_I2C_Write: 2, // 0x02
  Cmd_I2C_Read: 3, // 0x03
  Cmd_I2CRead_WithAddress: 4, // 0x04
  Cmd_SPI_Control: 8, // 0x08
  Cmd_SPI_WriteAndRead: 9, // 0x09
  Cmd_I2C_BlkWriteBlkRead: 96, // 0x60
  Cmd_I2C_ReadWithAddress: 100, // 0x64
  Cmd_I2C_ReadInternal: 101, // 0x65
  Cmd_I2C_WriteInternal: 102, // 0x66
  Cmd_SPI_StreamOut: 133, // 0x85
  Cmd_SPI_StreamStop: 134, // 0x86
  Cmd_SPI_WriteAndReadEx: 135, // 0x87
  Cmd_Pattern_Config: 159,
  Cmd_Pattern_Start: 160,
  Cmd_Pattern_Stop: 161,
  Cmd_Pattern_GetStatus: 162
};

let initialize = function() {
  /** *
   * I2C Interface sub-module definition
   ***/
  let I2C_100kHz = 0;
  let I2C_400kHz = 1;
  let I2C_10kHz = 2;
  let I2C_800kHz = 3;

  let I2C_7Bits = 0;
  let I2C_10Bits = 1;

  let I2C_PullUps_OFF = 0;
  let I2C_PullUps_ON = 1;

  let Pattern_Interface = function(settings, registerModel, u2a) {
    // console.log('Custom I2C', 'Constructor', settings, registerModel, u2a);
    this._registerModel = registerModel;
    this.cmdQueue = [];
    this.deviceAddrs = settings.deviceAddrs;
    this.u2a = u2a;

    this.configPattern = configPattern;
    this.startPattern = startPattern;
    this.stopPattern = stopPattern;
    this.getPatternStatus = getPatternStatus;
  };

  gc.databind.USB2ANY.prototype.PATTERN_Interface = Pattern_Interface;

  Pattern_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

  let configPattern = function(u2a, payload) {
    return u2a.sendCommandPacket(Command.Cmd_Pattern_Config, payload);
  };

  let startPattern = function(u2a, payload) {
    return u2a.sendCommandPacket(Command.Cmd_Pattern_Start, payload);
  };

  let stopPattern = function(u2a) {
    return u2a.sendCommandPacket(Command.Cmd_Pattern_Stop, []);
  };

  let getPatternStatus = function(u2a) {
    return u2a.sendCommandPacket(Command.Cmd_Pattern_GetStatus);
  };

  let CustomI2C_Interface = function(settings, registerModel, u2a) {
    // console.log('Custom I2C', 'Constructor', settings, registerModel, u2a);
    this._registerModel = registerModel;
    var i2cAddressHi = (settings.deviceAddrs >> 8) & 0xff;
    var i2cAddressLo = settings.deviceAddrs & 0xff;
    this.readyToProcess = true;
    this.cmdQueue = [];

    this.readData = [i2cAddressHi, i2cAddressLo];
    this.writeData = [i2cAddressHi, i2cAddressLo];

    this.sequentialRead = settings.sequentialRead;
    this.readWithAddress = settings.readWithAddress;

    this.u2a = u2a;

    if (settings.crc) {
      this.crc = new gc.databind.CRC(settings.crc);
    }
    if (settings.dataEndian === "little") {
      this._setBytes = setBytesLSB;
      this._getResult = getResultLSB;
      this._dataEndian = settings.dataEndian;
    } else {
      this._setBytes = setBytes;
      this._getResult = getResult;
    }
  };

  gc.databind.USB2ANY.prototype.maxOutstandingCommands = 30;

  /**
   * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
   * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
   * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
   */
  gc.databind.USB2ANY.prototype.CUSTOMI2C_Interface = CustomI2C_Interface;

  /**
   * Custom interfaces should be extended from the ICommInterface base class
   */
  CustomI2C_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

  CustomI2C_Interface.prototype.control = function(u2a, settings) {
    let speed = I2C_10kHz;
    switch (settings.speed) {
      case 100:
        speed = I2C_100kHz;
        break;
      case 400:
        speed = I2C_400kHz;
        break;
      case 800:
        speed = I2C_800kHz;
        break;
    }

    let addrsBits = I2C_7Bits;
    switch (settings.addrsBits) {
      case 10:
        addrsBits = I2C_10Bits;
        break;
    }

    let internalAddrsBits = parseNumberProperty(
      "internalAddrsBits",
      settings.internalAddrsBits || 8,
      1,
      16
    );
    this.internalAddrsBytes = Math.ceil(internalAddrsBits / 8);

    let pullUps = settings.pullup ? I2C_PullUps_ON : I2C_PullUps_OFF;

    u2a.sendCommandPacket(Command.Cmd_I2C_Control, [speed, addrsBits, pullUps]);
  };

  calculate8BitRegAddr = function(registerAddress) {
    // Remove MSB bits of register address from 10 bit register address
    let binaryValueOfRegAddr = (
      "00000000" + registerAddress.toString(2)
    ).substr(-10);

    let regAddr8Bit = binaryValueOfRegAddr.slice(2); // msbRegAdd = "01"

    return parseInt(regAddr8Bit, 2);
  };

  CustomI2C_Interface.prototype.read = function(
    u2a,
    info,
    registerModel,
    coreIndex
  ) {
    setDeviceAddress(this._registerModel, this.readData, info);
    let numBytes =
      info.nBytes !== undefined
        ? info.nBytes
        : Math.ceil(info.size !== undefined ? info.size / 8 : 1);

    if (this.readWithAddress) {
      // Using I2C ReadWithAddress API
      this.readData[2] = calculate8BitRegAddr(info.addr) || 0;
      this.readData[3] = numBytes;
    } else {
      // Using I2C ReadInternal API
      // readData[0-1] - device address
      // readData[2] - size of internal address, in bytes (must be 0, 1, or 2)
      this.readData[2] = this.internalAddrsBytes;

      // readData[3-4] - number of bytes of data
      this.readData[3] = (numBytes >> 8) & 0xff;
      this.readData[4] = numBytes & 0xff;

      // readData[5-6] - Internal address of the data to read
      if (this.internalAddrsBytes == 1) {
        // 1 byte register address
        this.readData[5] = calculate8BitRegAddr(info.addr) || 0;
      } else {
        // 2 byte register address
        this.readData[5] = calculate8BitRegAddr(info.addr)
          ? (calculate8BitRegAddr(info.addr) >> 8) & 0xff
          : 0;
        this.readData[6] = calculate8BitRegAddr(info.addr)
          ? calculate(info.addr) & 0xff
          : 0;
      }
    }

    let cmd = this.readWithAddress
      ? Command.Cmd_I2C_ReadWithAddress
      : Command.Cmd_I2C_ReadInternal;

    if (this.crcUser && this.crc) {
      let data = this.readData.slice();
      let crcData = this.crcUser.embed_crc_data(this.crc, {
        write: false,
        deviceAddr: (this.readData[0] << 8) | (this.readData[1] & 0xff),
        registerAddr: calculate8BitRegAddr(info.addr),
        payload: data,
        numBytes: numBytes
      });

      data = (crcData && crcData.payload) || data;

      if (crcData && crcData.numBytes) {
        if (this.readWithAddress) {
          data[3] = crcData.numBytes;
        } else {
          data[3] = (crcData.numBytes >> 8) & 0xff;
          data[4] = crcData.numBytes & 0xff;
        }
      }
      let self = this;
      return u2a
        .readResponse(u2a.sendCommandPacket(cmd, data))
        .then(
          function(result) {
            let payload = getPayload(result);
            let crcData = this.crcUser.verify_crc_data(this.crc, {
              payload: payload
            });

            if (crcData && crcData.valid === false) {
              return Q.reject("Invalid CRC");
            }

            return (crcData && crcData.payload) || payload;
          }.bind(this)
        )
        .then(function(array) {
          let result = 0;
          if (self._dataEndian === "little") {
            for (var i = array.length - 1; i >= 0; i--) {
              result = (result << 8) | (array[i] & 0xff);
            }
          } else {
            for (var i = 0; i < array.length; i++) {
              result = (result << 8) | (array[i] & 0xff);
            }
          }
          return result;
        });
    } else {
      return u2a
        .readResponse(u2a.sendCommandPacket(cmd, this.readData))
        .then(this._getResult);
    }
  };

  let MAX_COUNT = MAX_PAYLOAD - 6;

  CustomI2C_Interface.prototype.multiRegisterReadOnePacketHelper = function(
    u2a,
    startAddrs,
    regCount,
    numBytesPerRegister,
    registerModel,
    coreIndex
  ) {
    var cmd = Command.Cmd_I2C_ReadWithAddress;
    var numReadBytes = regCount * numBytesPerRegister;
    if (this.readWithAddress) {
      this.readData[2] = startAddrs;
      this.readData[3] = numReadBytes;
    } else if (this.blockWriteBlockRead) {
      cmd = Command.Cmd_I2C_BlkWriteBlkRead;
      this.readData[2] = this.internalAddrsBytes; // number of write bytes
      this.readData[3] = numReadBytes & 0xff; // number of read bytes
      // bytes to write is register address in this case
      if (this.internalAddrsBytes === 1) {
        this.readData[4] = startAddrs;
      } else {
        this.readData[4] = (startAddrs >> 8) & 0xff;
        this.readData[5] = startAddrs & 0xff;
      }
    } else {
      cmd = Command.Cmd_I2C_ReadInternal;
      // readData[2] - size of internal address, in bytes (must be 0, 1, or 2)
      this.readData[2] = this.internalAddrsBytes;
      // readData[3-4] - number of bytes of data
      this.readData[3] = (numReadBytes >> 8) & 0xff;
      this.readData[4] = numReadBytes & 0xff;
      // readData[5-6] - Internal address of the data to read
      if (this.internalAddrsBytes === 1) {
        // 1 byte register address
        this.readData[5] = startAddrs;
      } else {
        // 2 byte register address
        this.readData[5] = (startAddrs >> 8) & 0xff;
        this.readData[6] = startAddrs & 0xff;
      }
    }
    return u2a
      .readResponse(u2a.sendCommandPacket(cmd, this.readData))
      .then(getPayload)
      .then(payload => {
        var array = [];
        if (this._dataEndian === "little") {
          for (var reg = 0; reg < regCount; reg++) {
            var result = 0;
            for (var i = numBytesPerRegister - 1; i >= 0; i--) {
              result =
                (result << 8) | (payload[reg * numBytesPerRegister + i] & 0xff);
            }
            array[reg] = result;
          }
        } else {
          for (var reg = 0; reg < regCount; reg++) {
            var result = 0;
            for (var i = 0; i < numBytesPerRegister; i++) {
              result =
                (result << 8) | (payload[reg * numBytesPerRegister + i] & 0xff);
            }
            array[reg] = result;
          }
        }
        return array;
      });
  };

  CustomI2C_Interface.prototype.multiRegisterRead = function(
    u2a,
    startRegisterInfo,
    count,
    registerModel,
    coreIndex
  ) {
    registerModel = registerModel || this._registerModel;
    setDeviceAddress(registerModel, this.readData, startRegisterInfo);
    var numBytesPerRegister =
      startRegisterInfo.nBytes !== undefined
        ? startRegisterInfo.nBytes
        : Math.ceil(
            startRegisterInfo.size !== undefined
              ? startRegisterInfo.size / 8
              : 1
          );

    // BTW, for Cmd_I2C_BlkWriteBlkRead, the c code says 'if numWriteBytes + numReadBytes > 32, return error 'too much data'
    // This is much fewer than MAX_COUNT (48) and MAX_PAYLOAD (54).
    // If the cases, which expect to use Cmd_I2C_BlkWriteBlkRead, work well with Cmd_I2C_ReadInternal,
    // I suggest to drop the support of Cmd_I2C_BlkWriteBlkRead, unless we code in the limit of 32 for this special case.
    var maxRegisterCountPerPacket = Math.floor(MAX_COUNT / numBytesPerRegister);
    var startAddrs = startRegisterInfo.addr || 0;
    if (count > maxRegisterCountPerPacket) {
      // need to split into two or more hid packets.
      var promises = [];

      while (count > maxRegisterCountPerPacket) {
        promises.push(
          this.multiRegisterReadOnePacketHelper(
            u2a,
            calculate8BitRegAddr(startAddrs),
            maxRegisterCountPerPacket,
            numBytesPerRegister,
            registerModel,
            coreIndex
          )
        );
        setDeviceAddress(
          registerModel,
          this.readData,
          startRegisterInfo,
          startAddrs
        );
        count -= MAX_COUNT;
        startAddrs += MAX_COUNT;
      }
      setDeviceAddress(
        registerModel,
        this.readData,
        startRegisterInfo,
        startAddrs
      );
      promises.push(
        this.multiRegisterReadOnePacketHelper(
          u2a,
          calculate8BitRegAddr(startAddrs),
          count,
          numBytesPerRegister,
          registerModel,
          coreIndex
        )
      );
      return Q.all(promises).then(concatenateResults);
    } else {
      return this.multiRegisterReadOnePacketHelper(
        u2a,
        calculate8BitRegAddr(startAddrs),
        count,
        numBytesPerRegister,
        registerModel,
        coreIndex
      );
    }
  };

  let writePromise = Q(true);

  CustomI2C_Interface.prototype.write = function(
    u2a,
    info,
    value,
    registerModel,
    coreIndex
  ) {
    let nBytes =
      info.nBytes !== undefined
        ? info.nBytes
        : Math.ceil(info.size !== undefined ? info.size / 8 : 1);
    let size = 4 + nBytes;
    if (size <= this.writeData.length) {
      this.writeData = this.writeData.slice(0, size);
    }
    registerModel = registerModel || this._registerModel;
    setDeviceAddress(registerModel, this.writeData, info);

    this.writeData[2] = nBytes + 1;
    this.writeData[3] =
      (info.writeAddr === undefined
        ? calculate8BitRegAddr(info.addr)
        : info.writeAddr) || 0;
    this._setBytes(this.writeData, nBytes, value, 4);

    console.log("REG:WRITE: ", this.writeData[2], " value : ", value);

    let data = this.writeData.slice();
    if (this.crcUser && this.crc) {
      let crcData = this.crcUser.embed_crc_data(this.crc, {
        write: true,
        deviceAddr: (this.writeData[0] << 8) | (this.writeData[1] & 0xff),
        registerAddr: this.writeData[3],
        writeData: value,
        payload: data,
        numBytes: nBytes
      });

      data = (crcData && crcData.payload) || data;

      if (crcData && crcData.numBytes) {
        data[2] = Math.min(crcData.numBytes + 1, data.length - 3); // GC-2381
      }
    }

    u2a.sendCommandPacket(Command.Cmd_I2C_Write, data);
    return writePromise;
  };

  CustomI2C_Interface.prototype.initSymbolsForDevice = function(
    settings,
    registerModel
  ) {
    registerModel.readDeviceAddrsMap("I2C", settings);
  };

  CustomI2C_Interface.prototype.I2C_Read = function(slaveAddr, numBytes) {
    let buffer = [(slaveAddr >> 8) & 0xff, slaveAddr & 0xff, numBytes & 0xff];
    return this.u2a
      .readResponse(this.u2a.sendCommandPacket(Command.Cmd_I2C_Read, buffer))
      .then(getPayload);
  };

  CustomI2C_Interface.prototype.I2C_Write = function(slaveAddr, pDataBuffer) {
    pDataBuffer = pDataBuffer instanceof Array ? pDataBuffer : [pDataBuffer];
    let buffer = [
      (slaveAddr >> 8) & 0xff,
      slaveAddr & 0xff,
      pDataBuffer.length & 0xff
    ];
    buffer.push.apply(buffer, pDataBuffer);
    return this.u2a.readResponse(
      this.u2a.sendCommandPacket(Command.Cmd_I2C_Write, buffer)
    );
  };

  /** *
   * SPI Interface sub-module definition
   ***/

  let CustomSPI_Interface = function(settings, registerModel, u2a) {
    // console.log('Custom I2C', 'Constructor', settings, registerModel, u2a);
    var dataBits = settings.dataBits !== undefined ? settings.dataBits : 8;
    this.dataBitsMask = (1 << dataBits) - 1;
    this.dataBitsOffset = settings.dataBitsOffset || 0;

    var addrsBits = settings.addrsBits !== undefined ? settings.addrsBits : 6;
    this.addrsBitsMask = (1 << addrsBits) - 1;
    this.addrsBitsOffset = settings.addrsBitsOffset || 8;

    var parityMap = {even: 0, odd: 1};
    if (settings.parity) {
      this.parity = parseStringProperty(
        "parity",
        settings.parity || "even",
        parityMap
      );
    }
    this.parityBitsOffset = settings.parityBitsOffset || 0;

    var bitSize;
    this.writeFlag = 0;
    this.readFlag = 0;
    if (settings.readBitOffset !== undefined) {
      bitSize = settings.readBitOffset;
      this.readFlag = 1 << bitSize;
    } else if (settings.writeBitOffset !== undefined) {
      bitSize = settings.writeBitOffset;
      this.writeFlag = 1 << bitSize;
    } else {
      bitSize = 14;
      this.writeFlag = 1 << bitSize;
    }

    this.readWriteData = [0];
    bitSize = Math.max(
      bitSize,
      addrsBits + this.addrsBitsOffset,
      dataBits + this.dataBitsOffset,
      this.parityBitsOffset
    );
    var size = 0;
    for (var i = 0; i < bitSize; i += 8) {
      this.readWriteData.push(0);
      size++;
    }
    this.readWriteData[0] = size;

    console.log("Size : ", size);

    var shift = this.dataBitsOffset;
    var mask = this.dataBitsMask;
    var parity = this.parity;

    this._getResult = function(data) {
      var command = getResult(data);
      if (parity !== undefined && parity !== computeParity(command, size * 8)) {
        throw "parity error on register data received";
      }
      return (command >>> shift) & mask;
    };
  };

  /**
   * Add the custom interface to the USB2ANY prototype, the name must take the form of "XXX_Interface" where "XXX" is all capitalized, eg. "LED_Interface"
   * To use custom interfaces in the system.json, set the name property to be the "XXX" part of the interface name (not case-sensitive), eg. "name" : "led"
   * To avoid naming conflicts with already defined interfaces, do not use names "power", "i2c", "spi", "gpio", "uart", "pwm" or "displayscale"
   */
  gc.databind.USB2ANY.prototype.CUSTOMSPI_Interface = CustomSPI_Interface;

  /**
   * Custom interfaces should be extended from the ICommInterface base class
   */
  CustomSPI_Interface.prototype = new gc.databind.USB2ANY.ICommInterface();

  CustomSPI_Interface.prototype.control = function(u2a, settings) {
    var clockPhaseMap = {following: 0, first: 1};
    var clockPolarityMap = {low: 0, high: 1};
    var bitDirectionMap = {lsb: 0, msb: 1};
    var latchTypeMap = {
      byte: 0,
      packet: 1,
      word: 2,
      no_cs: 3,
      pulse_after_packet: 255
    };
    var latchPolarityMap = {high: 0, low: 1};

    var clockPhase = parseStringProperty(
      "clockPhase",
      settings.clockPhase || "first",
      clockPhaseMap
    );
    var clockPolarity = parseStringProperty(
      "clockPolarity",
      settings.clockPolarity || "low",
      clockPolarityMap
    );
    var bitDirection = parseStringProperty(
      "bitDirection",
      settings.bitDirection || "MSB",
      bitDirectionMap
    );
    var characterLength =
      8 -
      parseNumberProperty(
        "characterLength",
        settings.characterLength || 8,
        7,
        8
      );
    var latchType = parseStringProperty(
      "latchType",
      settings.latchType || "packet",
      latchTypeMap
    );
    var latchPolarity = parseStringProperty(
      "latchPolarity",
      settings.latchPolarity || "low",
      latchPolarityMap
    );

    var divider = settings.clockDivider || 1;
    var dividerHigh = (divider >>> 8) & 0xff;
    var dividerLow = divider & 0xff;

    u2a.sendCommandPacket(Command.Cmd_SPI_Control, [
      clockPhase,
      clockPolarity,
      bitDirection,
      characterLength,
      latchType,
      latchPolarity,
      dividerHigh,
      dividerLow
    ]);
  };

  CustomSPI_Interface.prototype.read = function(
    u2a,
    info,
    registerModel,
    coreIndex
  ) {
    var command =
      this.readFlag |
      (((info.addr || 0) & this.addrsBitsMask) << this.addrsBitsOffset);
    if (this.parity !== undefined) {
      var parity =
        this.parity ^ computeParity(command, this.readWriteData[0] * 8);
      command = command | (parity << this.parityBitsOffset);
    }

    // setBytes(this.readWriteData, this.readWriteData[0], command, 1);

    this.readWriteData[0] = 3;
    this.readWriteData[1] = calculateAddressByte1(info.addr);
    this.readWriteData[2] = calculateAddressByte2(info, "read");
    this.readWriteData[3] = 0;

    return u2a
      .readResponse(
        u2a.sendCommandPacket(Command.Cmd_SPI_WriteAndRead, this.readWriteData)
      )
      .then(this._getResult);
  };

  calculateAddressByte2 = function(regInfo, readOrWrite) {
    let addr = regInfo.addr || 0;
    let msbByte = new Array(8);
    binaryValueOfRegAddr = ("0000000000" + addr.toString(2)).substr(-10);
    let msbBit = binaryValueOfRegAddr.slice(8, 10);
    msbBit = msbBit.split("").map(Number);

    msbByte[0] = msbBit[0];
    msbByte[1] = msbBit[1];
    msbByte[2] = readOrWrite === "read" ? 0 : 1;
    msbByte.push(...new Array(5).fill(0));
    return parseInt(msbByte.join(""), 2);
  };

  calculateAddressByte1 = function(address) {
    let addr = address || 0;
    binaryValueOfRegAddr = ("0000000000" + addr.toString(2)).substr(-10);
    let msbBit = binaryValueOfRegAddr.slice(0, 8);

    msbBit = msbBit.split("").map(Number);
    return parseInt(msbBit.join(""), 2);
  };

  CustomSPI_Interface.prototype.write = function(
    u2a,
    info,
    value,
    registerModel,
    coreIndex
  ) {
    var command =
      this.writeFlag |
      (((info.addr || 0) & this.addrsBitsMask) << this.addrsBitsOffset) |
      ((value & this.dataBitsMask) << this.dataBitsOffset);

    if (this.parity !== undefined) {
      var parity =
        this.parity ^ computeParity(command, this.readWriteData[0] * 88);
      command = command | (parity << this.parityBitsOffset);
    }

    this.readWriteData[0] = 3;
    this.readWriteData[1] = calculateAddressByte1(info.addr);
    this.readWriteData[2] = calculateAddressByte2(info, "write");
    this.readWriteData[3] = value;

    // setBytes(this.readWriteData, this.readWriteData[0], command, 1);
    u2a.sendCommandPacket(Command.Cmd_SPI_WriteAndRead, this.readWriteData);
    return writePromise;
  };

  /** *
   * Helper functions for USB2ANY sub-modules, like I2C, Power, GPIO, and SPI
   ***/

  let setDeviceAddress = function(
    registerModel,
    buffer,
    info,
    startAddrs = false
  ) {
    let deviceAddrs;
    if (info.deviceAddrs && !isNaN(info.deviceAddrs)) {
      deviceAddrs = +info.deviceAddrs;
    } else if (registerModel) {
      deviceAddrs = registerModel.getDeviceAddrsForRegister(info);
    }

    buffer[0] = (deviceAddrs >> 8) & 0xff;
    buffer[1] = deviceAddrs & 0xff;
  };

  let setBytesLSB = function(array, size, value, offset) {
    // little endian
    for (let i = 0; i < size; i++) {
      array[offset + i] = value & 0xff;
      value = value >>> 8;
    }
  };

  let setBytes = function(array, size, value, offset) {
    for (let i = size; i-- > 0; ) {
      array[offset + i] = value & 0xff;
      value = value >>> 8;
    }
  };

  let concatenateResults = function(results) {
    return Array.prototype.concat.apply([], results);
  };

  let getMultipleRegisterReadResult = function(array) {
    return array.slice(PACKET_PAYLOAD);
  };

  let getResult = function(array) {
    let result = 0;
    let size = PACKET_PAYLOAD + (array[PACKET_PAYLOAD_LEN] || 0);
    for (let i = PACKET_PAYLOAD; i < size; i++) {
      result = (result << 8) | (array[i] & 0xff);
    }
    return result;
  };

  let getResultLSB = function(array) {
    // little endian
    let result = 0;
    let size = PACKET_PAYLOAD + (array[PACKET_PAYLOAD_LEN] || 0);
    for (let i = size - 1; i >= PACKET_PAYLOAD; i--) {
      result = (result << 8) | (array[i] & 0xff);
    }
    return result;
  };

  let getPayload = function(array) {
    return array.slice(
      PACKET_PAYLOAD,
      PACKET_PAYLOAD + (array[PACKET_PAYLOAD_LEN] || 0)
    );
  };

  let TheValue = 'The value "';
  let ForEntry = '", for entry ';
  let InRegDefs = ", in the system.json file";

  let parseNumberProperty = function(message, value, min, max) {
    let result = +value;
    if (isNaN(result)) {
      throw TheValue +
        value +
        ForEntry +
        message +
        InRegDefs +
        " is not a number.";
    }
    if (min && result < min) {
      throw TheValue +
        value +
        ForEntry +
        message +
        InRegDefs +
        " must be greater than " +
        min +
        ".";
    }
    if (max && result > max) {
      throw TheValue +
        value +
        ForEntry +
        message +
        InRegDefs +
        " must be less than " +
        max +
        ".";
    }
    return result;
  };

  let parseStringProperty = function(message, value, valueMap) {
    let stringValue = ("" + value).toLowerCase();
    if (valueMap.hasOwnProperty(stringValue)) {
      return valueMap[stringValue];
    } else {
      message =
        TheValue +
        value +
        ForEntry +
        message +
        InRegDefs +
        " is not supported.  Valid entries are";
      let delimiter = ' "';
      let lastOption;
      for (let option in valueMap) {
        if (valueMap.hasOwnProperty(option)) {
          if (lastOption) {
            message = message + delimiter + lastOption;
            delimiter = '", "';
          }
          lastOption = option;
        }
      }
      throw message + '", or "' + lastOption + '".';
    }
  };
};

// Wait until USB2ANY global var is ready before defining interface
let template = document.querySelector("#template_obj");
if (template) {
  initialize();
} else {
  if (gc.desktop.isNodeWebkit()) {
    document.addEventListener("WebComponentsReady", initialize);
  } else {
    document.addEventListener("WebComponentsReady", initialize);
  }
}
